from phdi.fhir.linkage.link import add_patient_identifier
from phdi.fhir.linkage.link import add_patient_identifier_in_bundle

__all__ = ["add_patient_identifier_in_bundle", "add_patient_identifier"]
